import os, json, jmespath, copy

from yac.lib.registry import get_registry_keys, get_remote_value, clear_entry_w_challenge
from yac.lib.registry import set_remote_string_w_challenge
from yac.lib.file import get_file_contents
from yac.lib.variables import get_variable, set_variable
from yac.lib.validator import validate_dictionary

NULL_PARAMS = "service params do not exist"

INVALID_PARAMS = "service params are not valid"

SERVICE_PARAMS_SUFFIX="-service-params"

REQUIRED_FIELDS = ["params-name.value",
                   "params"]

SUPPORTED_KEYS = ["params-name",
                  "params",
                  "description"]

class ServiceError():
    def __init__(self, msg):
        self.msg = msg

def validate_service_params(service_params):

    # validate the main services
    val_errors = validate_dictionary(service_params, SUPPORTED_KEYS, REQUIRED_FIELDS)
    
    return val_errors


def get_all_service_params_names(search_string=""):

    params_names = []

    # get all registry keys
    registry_keys = get_registry_keys()

    # find all keys with _naming suffix
    for key in registry_keys:

        if SERVICE_PARAMS_SUFFIX in key:
            # remove the suffix and append to list
            
            if search_string and search_string in key:
                params_names = params_names + [key.replace(SERVICE_PARAMS_SUFFIX,'')]

    return params_names   

# determine service alias, service key, and service defintion based on the stack cli args
# second argument allows this fxn to be called recursively in support of nested services
def get_service_params(serviceparams_arg):

    # treat the arg as a path to a local file
    service_params, params_name = get_service_params_from_file(serviceparams_arg)

    if (serviceparams_arg and params_name == NULL_PARAMS):

        # Service params do not exist as a local file

        # Treat serviceparams_arg as the service name. See if it is complete (i.e.
        # includes a version label)
        if is_service_params_name_complete(serviceparams_arg):

            # name is complete
            params_name = serviceparams_arg

        # Treat serviceparams_arg is a service name that lacks a version
        elif is_service_params_available_partial_name(serviceparams_arg):

            # get complete name
            params_name = get_complete_name(serviceparams_arg)

        # pull the service from the registry
        service_params, params_name = get_service_params_by_name(params_name)

    return service_params, params_name

def get_service_params_from_file(serviceparams_arg, params_only=True):

    service_params = {}
    params_name = NULL_PARAMS
    abs_path = ""

    if serviceparams_arg:

        file_contents = get_file_contents(serviceparams_arg)

        # pull the service params from file
        if file_contents:

            service_params_descriptor = json.loads(file_contents)

            # do a quick validation of the params
            if not validate_service_params(service_params_descriptor):

                # pull the service name out of the descriptor
                params_name = get_variable(service_params_descriptor, 'params-name')      

                if params_only:
                    # pull the actual parameters out for return
                    service_params = service_params_descriptor['params']
                else:
                    # return the entire contents of the descriptor
                    service_params = service_params_descriptor

            else:
                params_name = INVALID_PARAMS

    return service_params, params_name

def get_service_params_by_name(params_name):

    service_params = {}

    if params_name:

        reg_key = params_name + SERVICE_PARAMS_SUFFIX

        # look in remote registry
        service_contents = get_remote_value(reg_key)

        if service_contents:

            # load into dictionary
            service_params_descriptor = json.loads(service_contents)

            # pull the actual parameters out for return
            service_params = service_params_descriptor['params']

        else:
            # service doesn't exist
            params_name = NULL_PARAMS            

    return service_params, params_name


def clear_service_params(params_name, challenge):

    # if service is in fact registered
    service_params, params_name_returned = get_service_params_by_name(params_name)
    if service_params:

        # clear service entry registry
        reg_key = params_name + SERVICE_PARAMS_SUFFIX
        clear_entry_w_challenge(reg_key, challenge)     
    
    else:
        raise ServiceError("service params with key %s doesn't exist"%params_name)

# register service into yac registry
def register_service_params(params_name, service_path, challenge):

    if os.path.exists(service_path):

        service_contents_str = get_file_contents(service_path)

        if service_contents_str:

            reg_key = params_name + SERVICE_PARAMS_SUFFIX

            # set the service in the registry
            set_remote_string_w_challenge(reg_key, service_contents_str, challenge)

    else:
        raise ServiceError("service path %s doesn't exist"%service_path)
 

# a service name is considered complete if it includes a version tag
def is_service_params_name_complete(params_name):

    is_complete = False

    name_parts = params_name.split(':')

    if len(name_parts)==2:

        # a tag is included, so name is complete
        is_complete = True

    return is_complete  

# if only know partial service name (no label), returns true
# if the complete version of the service is in registry
def is_service_params_available_partial_name(service_partial_name):

    is_available = False

    if not is_service_params_name_complete(service_partial_name):
        # see if a service with tag=latest is available in the registry
        complete_name_candidate = '%s:%s'%(service_partial_name,"latest")
        service_desc, params_name = get_service_params_by_name(complete_name_candidate)

        if service_desc:
            is_available = True

    return is_available

def get_complete_name(params_name):

    complete_name = ""

    if not is_service_params_name_complete(params_name):
        # see if a service with tag=latest is available in the registry
        complete_name_candidate = '%s:%s'%(params_name,"latest")
        service_desc, params_name = get_service_params_by_name(complete_name_candidate)

        if service_desc:
            complete_name = complete_name_candidate

    return complete_name






#!/usr/bin/env python

import argparse, sys, json, os

from yac.lib.registry import RegError
from yac.lib.params import validate_service_params, register_service_params, clear_service_params, get_service_params_by_name
from yac.lib.params import get_service_params_from_file, get_all_service_params_names, NULL_PARAMS

def is_valid_file(parser, arg):
    if not os.path.exists(arg):
        parser.error("The file %s does not exist!" % arg)
    else:
        return arg

def main():

    parser = argparse.ArgumentParser('Install a set of service params in the yac registry')
    group = parser.add_argument_group()
    group.add_argument('--share', 
                        help=('register a set  of service params with the yac registry so they can be easily consumed by others.' +
                        ' (arg is the path to the params file)'),
                        type=lambda x: is_valid_file(parser, x))
    group.add_argument('--ver',  help='the version label to associate with this set of params')
    parser.add_argument('--show',  help='show a set of params (arg is the params key)')    
    parser.add_argument('--clear', help='clear a set of params from the registry  (arg is the service key)')
    parser.add_argument('--find',  help='find a set of params in the registry via a search string ',
                                   default="") 

    # pull out args
    args = parser.parse_args()


    if args.share:

        # pull descriptor from file
        service_params, params_name_partial = get_service_params_from_file(args.share,False)

        if not args.ver:
            print "Please specify a version label via the --ver argument"
            exit()

        params_name = "%s:%s"%(params_name_partial,args.ver)

        # validate the service description
        validation_errors = validate_service_params(service_params)

        if not validation_errors:

            # see if service has already been registered
            service_params_in_registry, params_name_in_registry = get_service_params_by_name(params_name)

            if not service_params_in_registry:
                challenge = raw_input("Please input a challenge phrase to control updates to your service params >> ")
            else:
                challenge = raw_input("Please input the challenge phrase associated with this set of service params >> ")


            print ("About to register params '%s' with challenge phrase '%s'. ")%(params_name,challenge)
            raw_input("Hit Enter to continue >> ")

            try:
                register_service_params(params_name, args.share, challenge)

                print ("Your params have been registered with yac under the key: '%s' and challenge phrase '%s'.\n" +
                       "You will prompted for the challenge phrase if you attempt to update your params.\n" + 
                       "Other users can run your params via '>> yac stack <servicename> -p %s ...'")%(params_name,challenge,params_name)

            except RegError as e:
                print ("Your params registration failed: %s"%e.msg)

        else:
            print ("Your service params failed validation checks. Please fix the following errors. %s"%validation_errors)
        

    elif args.clear:

        print "Clearing the '%s' service from registry"%(args.clear)
        challenge = raw_input("Please input the challenge phrase associated with this service >> ")        
        raw_input("Hit Enter to continue >> ")

        clear_service_params(args.clear, challenge)

        # TODO: need some success/fail feedback

    elif args.find:
        all_params = get_all_service_params_names(args.find)
        print all_params

    elif args.show:
        service_params, params_name_in_registry = get_service_params_by_name(args.show)
        if params_name_in_registry != NULL_PARAMS:
            print json.dumps(service_params,indent=2)
        else:
            print "Params requested are not available in registry"      




#!/usr/bin/env python

import argparse, sys, boto3

from yac.lib.inputs import get_stack_service_inputs
from yac.lib.variables import get_variable
from yac.lib.service import get_service, NULL_SERVICE, get_service_parmeters, get_service_alias
from yac.lib.db import DB, rename_rds_instance, snapshot_exists, restore_intance_from_snapshot
from yac.lib.db import restore_security_group, delete_instance, instance_exists
from yac.lib.vpc import get_vpc_prefs
from yac.lib.naming import set_namer, get_resource_name

def prompt_for_confirmation(msg):

    sys.stdout.write('\r')                            
    sys.stdout.write( '%s%s%s'%('\033[92m',msg, '\033[0m'))
    sys.stdout.flush() 
    raw_input(" ")

def main():

    parser = argparse.ArgumentParser(description='Restore a DB from an RDS snapshot')

    # required args                                         
    parser.add_argument('instance', help='the id of the RDS instance to restore')
    parser.add_argument('snap', help='the name of the RDS snapshot to restore from')

    parser.add_argument('-d','--dryrun',  help='dry run the container start by printing rendered template to stdout', 
                                          action='store_true')

    parser.add_argument('-s','--skipto',  help='skip to this step number')

    # pull out args
    args = parser.parse_args()

    original_instance_id = args.instance

    # verify snapshot exists
    if not snapshot_exists(args.snap): 
        print "Snapshot specified does not exist. Please fix"
        exit()

    if args.dryrun:

        msg = ('\n(Dry-run) restore of the %s snapshot into the %s RDS instance.\n' +
                   'Hit <enter> to continue...')%(args.snap, original_instance_id)
        prompt_for_confirmation(msg)

    else:

        # rename the rds instance to clear the way for the restore

        # first define a new, "temp" name for the existing instance
        temp_instance_id = original_instance_id + '-temp'

        if not args.skipto or int(args.skipto) <= 1:

            if not instance_exists(original_instance_id):
                print "RDS instance specified does not exist. Please try again"
                exit()

            msg = ('About to restore the %s snapshot into the %s RDS instance \n' +
                   'Hit <enter> to continue...')%(args.snap, original_instance_id)
            prompt_for_confirmation(msg)  

            # rename the existing RDS instance
            rename_rds_instance(original_instance_id,temp_instance_id)

            msg =('(Step 1 of 3) The %s RDS instance has been renamed to %s in order to clear the way but must be rebooted before the restore can start.\n' +
                  'Hit <enter> when the %s instance shows as status=available on rds console ' +
                  ' (takes a few minutes or so) ...')%(original_instance_id,temp_instance_id,temp_instance_id)
            
            prompt_for_confirmation(msg)

        if args.skipto and int(args.skipto) <= 2:

            msg =('(Step 2 of 3) The %s snapshot can now be restored into the %s instance.\n' +
                  'Hit <enter> to proceed (takes 10 minutes or so ' +
                  ' depending on DB size) ...')%(args.snap, original_instance_id)
            prompt_for_confirmation(msg)

            restore_intance_from_snapshot(original_instance_id,temp_instance_id, args.snap)

        if args.skipto and int(args.skipto) <= 3:

            msg = ('(Step 3 of 4) The %s RDS instance has been restored from snapshot but must be rebooted before the security group can be re-applied.\n' +
                   'Hit <enter> when the instance shows as status=available on rds console (takes 15-20 minutes) ...')%(original_instance_id)
            prompt_for_confirmation(msg)

            restore_security_group(original_instance_id,original_instance_id)

        if args.skipto and int(args.skipto) <= 4:

            msg =('(Step 4 of 4) The %s RDS instance can now be deleted to complete the restore.\n' +
                  'Hit <enter> when the instance shows as status=available on rds console (takes few minutes) ...')%(temp_instance_id)
            prompt_for_confirmation(msg)

            delete_instance(temp_instance_id)

        print "Restore from snapshot is complete"


#!/usr/bin/env python

import argparse, os, inspect, jmespath, json
from colorama import Fore, Style
from yac.lib.service import get_service, get_service_parmeters, get_service_alias
from yac.lib.stack import get_stack_templates, get_stack_name, get_ec2_ips, deploy_stack_files
from yac.lib.naming import set_namer
from yac.lib.task import get_task, get_task_names, run_task
from yac.lib.params import get_service_params, NULL_PARAMS, INVALID_PARAMS
from yac.lib.vpc import get_vpc_prefs 
from yac.lib.variables import get_variable,set_variable

def main():

    parser = argparse.ArgumentParser(description='Run a task per the task defintions in the provided Servicefile')

    # required args                                         
    parser.add_argument('servicefile', help='location of the Servicefile (registry key or abspath)')
    parser.add_argument('-n','--name', help='name of task to run') 

    parser.add_argument('-p',
                        '--params',  help='path to a file containing additional, static, service parameters (e.g. vpc params, of service constants)') 
    parser.add_argument('-a',
                        '--alias',  help='service alias for the stack currently supporting this service (deault alias is per Servicefile)')
    parser.add_argument('-s',
                        '--show',  help='show a service',
                                   action='store_true')      
    parser.add_argument('--public',     help='connect using public IP address', 
                                        action='store_true')     

    args = parser.parse_args()

     # determine service defintion, complete service name, and service alias based on the args
    service_descriptor, service_name, servicefile_path = get_service(args.servicefile) 

    # abort if service_descriptor was not loaded successfully
    if not service_descriptor:
        print("The Servicefile input does not exist locally or in registry. Please try again.")
        exit()

    # get vpc preferences in place
    vpc_prefs = get_vpc_prefs()

    # set the resource namer to use with this service
    set_namer(service_descriptor, servicefile_path, vpc_prefs)

    # get the alias to use with this service
    service_alias = get_service_alias(service_descriptor,args.alias)

    # determine service params based on the params arg
    service_params_input, service_params_name = get_service_params(args.params)

    # abort if service params were not loaded successfully
    if args.params and service_params_name == NULL_PARAMS:
        print("The service params specified do not exist locally or in registry. Please try again.")
        exit()
    elif service_params_name == INVALID_PARAMS:
        print("The service params file specified failed validation checks. Please try again.")
        exit()
        
    # get the service parameters for use with yac-ref's in service templates
    service_parmeters = get_service_parmeters(service_alias, service_params_input, 
                                              service_name, service_descriptor,
                                              servicefile_path, vpc_prefs)

    # get cloud formation template for the service requested and apply yac intrinsic 
    # functions (yac-ref, etc.) using  the service_parmeters
    stack_template = get_stack_templates(service_descriptor,  
                                         service_parmeters)

    if args.show:

        print(Fore.GREEN)

        tasks = get_variable(service_parmeters,"tasks",{})

        if tasks:
            print "Available Tasks:"
            
            for task_name in tasks.keys():
                print "%s: %s"%(task_name,tasks[task_name]['comment'])
        else:
            print "There are no tasks associated with this service"

        print(Style.RESET_ALL)                

    elif args.name:

        # get the task params
        task_def = get_task(args.name, service_descriptor)

        if not task_def:

            task_names = get_task_names(service_parmeters)

            print("The task requested doesn't exist in the Servicefile. " +
                  "Available tasks include:\n%s."%pp_list(task_names))
            exit()

        print(Fore.GREEN)

        # run the requested task
        run_task(args.name,service_parmeters,stack_template)

        print(Style.RESET_ALL)

    else:

        print(Fore.RED)

        print("Nothing to do. Show available tasks using the --show argument or specify a task to run using the --name argument.")

        print(Style.RESET_ALL)        



def pp_list(list):
    str = ""
    for item in list:
        str = str + '* %s\n'%item

    return str


#!/usr/bin/env python

import argparse, jmespath, shutil, os
from yac.lib.naming import get_stack_name, set_namer
from yac.lib.service import get_service, get_service_parmeters, get_service_alias
from yac.lib.stack import get_stack_templates, get_stack_name, get_ec2_ips, deploy_stack_files
from yac.lib.vpc import get_vpc_prefs
from yac.lib.container.api import get_connection_str
from yac.lib.container.build import build_image, get_image_name, get_rendered_dockerpath
from yac.lib.template import apply_templates_in_dir
from yac.lib.container.config import get_aliases, find_image_tag
from yac.lib.params import get_service_params, NULL_PARAMS, INVALID_PARAMS

def main():
    
    parser = argparse.ArgumentParser(description='Build image for a service container unto a stack host.')     

    # required args   
    parser.add_argument('servicefile', help='location of the Servicefile (registry key or abspath)')
    parser.add_argument('name',        help='name of container to build')

    parser.add_argument('buildpath',   help='path to the dockerfile') 
        
    parser.add_argument('-d','--dryrun',  help='dry run the container start by printing rendered template to stdout', 
                                          action='store_true')   
    parser.add_argument('-a',
                        '--alias',  help='service alias for the stack currently supporting this service (deault alias is per Servicefile)')     
    parser.add_argument('-p',
                        '--params',  help='path to a file containing additional, static, service parameters (e.g. vpc params, of service constants)') 
    parser.add_argument('--public',     help='connect using public IP address', 
                                        action='store_true') 
                                            
    # pull out args
    args = parser.parse_args()

    # determine service defintion, complete service name, and service alias based on the args
    service_descriptor, service_name, servicefile_path = get_service(args.servicefile)

    # get vpc preferences in place
    vpc_prefs = get_vpc_prefs()

    # set the resource namer to use with this service
    set_namer(service_descriptor, servicefile_path, vpc_prefs)

    # get the alias to use with this service
    service_alias = get_service_alias(service_descriptor,args.alias)

    # determine service params based on the params arg
    service_params_input, service_params_name = get_service_params(args.params)

    # abort if service params were not loaded successfully
    if args.params and service_params_name == NULL_PARAMS:
        print("The service params specified do not exist locally or in registry. Please try again.")
        exit()
    elif service_params_name == INVALID_PARAMS:
        print("The service params file specified failed validation checks. Please try again.")
        exit()
        
    # get the service parameters for use with yac-ref's in service templates
    service_parmeters = get_service_parmeters(service_alias, service_params_input, 
                                              service_name, service_descriptor,
                                              servicefile_path, vpc_prefs)

    # get cloud formation template for the service requested and apply yac intrinsic 
    # functions (yac-ref, etc.) using  the service_parmeters
    stack_template = get_stack_templates(service_descriptor,
                                         service_parmeters)

    image_tag = find_image_tag(args.name,stack_template)

    # get stack name
    stack_name = get_stack_name(service_parmeters)

    # replace any service parmeters variables in body of any files
    build_path = get_rendered_dockerpath(image_tag) 
    if os.path.exists(build_path):
        shutil.rmtree(build_path)

    apply_templates_in_dir(args.buildpath, service_parmeters, build_path)

    # get the ip address of the target host
    target_host_ips = get_ec2_ips( stack_name , "", args.public)

    if args.dryrun:

        user_msg = "%s(dry-run) Building the %s container image on %s using the rendered dockerfile at %s%s"%('\033[92m',
                                                        image_tag,
                                                        target_host_ips,
                                                        build_path,                                                    
                                                        '\033[0m')
        print user_msg


    else:

        # build the container image on each host ...
        for target_host_ip in target_host_ips:

            user_msg = "%sBuilding the %s container image on %s using the rendered dockerfile at %s%s"%('\033[92m',
                                                        image_tag,
                                                        target_host_ip,
                                                        build_path,                                                    
                                                        '\033[0m')

            raw_input(user_msg + "\nHit <enter> to continue..." )

            # get connection string for the docker remote api on the target host
            docker_api_conn_str = get_connection_str( target_host_ip )

            # build the image
            build_image( image_tag, build_path, docker_api_conn_str )



#!/usr/bin/env python

import os, json
import jmespath

# xform env variables in taskdef format into a format compatible
# with docker api
def env_ec2_to_api(taskdef_envs):

    env_map = {}

    for env in taskdef_envs:
        env_map[env['Name']] = env['Value']

    return env_map

# xform volumes and mountpoints in taskdef format into a format compatible
# with docker api volume maps
def get_volumes_map(taskdef_volumes, taskdef_mountPoints):

    volume_map = {}
    volume_bindings = []
    for volume in taskdef_volumes:

        # example jmespath for 'home' volume is
        # [?sourceVolume=='home'].containerPath
        cp_search_str = "[?SourceVolume=='%s'].ContainerPath"%volume['Name']
        ro_search_str = "[?SourceVolume=='%s'].ReadOnly"%volume['Name']

        bind_path = jmespath.search(cp_search_str,taskdef_mountPoints)
        read_only = jmespath.search(ro_search_str,taskdef_mountPoints)

        # if we found a hit in the mount points, then add this volume to the 
        # volume map
        if (len(bind_path)==1 and len(read_only)==1):

            volume_bindings.append(volume['Host']['SourcePath'])
            
            volume_map[ volume['Host']['SourcePath'] ] = { 
                                    "bind": jmespath.search(cp_search_str,taskdef_mountPoints)[0],
                                    "ro":  jmespath.search(ro_search_str,taskdef_mountPoints)[0]
                                }


    return volume_map, volume_bindings

# xform port mapping in taskdef format into the port binding format compatible
# with docker api 
def get_port_bindings(taskdef_port_mappings):

    port_binding = {}

    for port_mapping in taskdef_port_mappings:
        port_binding[str(port_mapping['ContainerPort'])] = str(port_mapping['HostPort'])

    return port_binding


# taskdef_configs per ECS standard
def get_aliases(taskdef_configs):

    return jmespath.search('ContainerDefinitions[*].name', taskdef_configs)
  
def find_image_tag(name_tag, stack_template):

    image_tag = ""
    container_defs = get_container_defs(stack_template)

    image_tags = jmespath.search("[?Name=='%s'].Image"%name_tag,container_defs)

    # only one should be returned
    if len(image_tags)==1: 
        image_tag = image_tags[0]

    return image_tag

def get_container_envs(name_tag, stack_template):

    env_variables = []
    container_defs = get_container_defs(stack_template)

    ecs_env = jmespath.search("[?Name=='%s'].Environment"%name_tag,container_defs)

    if len(ecs_env) == 1:
        env_variables = env_ec2_to_api(ecs_env[0])

    return env_variables

def get_container_names(stack_template):

    container_names = []
    container_defs = get_container_defs(stack_template)

    container_names = jmespath.search("[*].Name",container_defs)

    return container_names

def get_container_volumes(name_tag, stack_template):

    app_taskdefs = get_task_defs(name_tag, stack_template)

    # print "%s"%json.dumps(app_taskdefs,indent=2)

    all_volumes = jmespath.search("Properties.Volumes",app_taskdefs)
    mount_points = jmespath.search("Properties.ContainerDefinitions[?Name=='%s'].MountPoints"%name_tag,app_taskdefs)[0]
    volumes_map, volumes_bindings = get_volumes_map(all_volumes,mount_points)

    return volumes_map, volumes_bindings

def get_container_ports(name_tag, stack_template):

    port_bindings = []
    container_defs = get_container_defs(stack_template)

    # get the port bindings for this container
    port_mappings = jmespath.search("[?Name=='%s'].PortMappings"%name_tag,container_defs)[0]
    port_bindings = get_port_bindings(port_mappings)

    return port_bindings

# get the task definitions for specific container
def get_task_defs(name_tag, stack_template):

    task_defs = {}

    if "Resources" in stack_template:

        resources = stack_template['Resources']

        resource_keys = resources.keys()

        for resource_key in resource_keys:

            if resources[resource_key]['Type'] == "AWS::ECS::TaskDefinition":

                container_defs = resources[resource_key]['Properties']['ContainerDefinitions']
                
                image_tags = jmespath.search("[?Name=='%s'].Image"%name_tag,container_defs)

                if len(image_tags)==1:

                    # this is the task definition of interest
                    task_defs = resources[resource_key]

    return task_defs 

def get_container_defs(stack_template):

    container_defs = []
    if "Resources" in stack_template:

        resources = stack_template['Resources']

        resource_keys = resources.keys()

        for resource_key in resource_keys:

            if resources[resource_key]['Type'] == "AWS::ECS::TaskDefinition":

                container_defs = container_defs + resources[resource_key]['Properties']['ContainerDefinitions']

    return container_defs 


#!/usr/bin/env python

from docker import Client
import docker, os, json, zipfile, urllib
from StringIO import StringIO
from yac.lib.container.api import get_docker_client
from yac.lib.template import apply_templates_in_dir

def build_image(
            image_tag,
            sources_path,
            connection_str = "",
            rm = "false"):

    print 'Getting docker client for connection_str=%s' % (connection_str)

    docker_client = get_docker_client( connection_str )

    # build the image
    for line in docker_client.build(tag=image_tag,
                                    path=sources_path,
                                    rm=rm):

        line_json = json.loads(line)

        if 'stream' in line_json:
            print line_json['stream']
        else:
            print line_json

def get_image_versions_local(connection_str, image_name):

    docker_client = get_docker_client( connection_str )

    # get the image
    images = docker_client.images()

    versions = []

    # print "local images: %s"%(images)

    for image in images:

        for tag in image['RepoTags']:

            if image_name in tag:

                versions = versions + [parse_image_version(tag)]
                break;

    return versions

def get_rendered_dockerpath(image_tag):

    # define a page where rendered dockerfiles can written

    home = os.path.expanduser("~")

    dockerfile_path = ""

    build_path = os.path.join(home,'.yac','dockerfiles', image_tag)

    return build_path 

def download_dockerfile(image_tag, repo_url, service_parmeters):

    home = os.path.expanduser("~")

    dockerfile_path = ""

    build_path = get_rendered_dockerpath(image_tag)

    tmp_path = os.path.join(home,'.yac',"tmp")

    name = os.path.join(tmp_path, 'temp.zip')
    try:
        name, hdrs = urllib.urlretrieve(repo_url, name)
    except IOError, e:
        print "Can't download %r to %r: %s" % (repo_url, name, e)
        return
    
    try:
        z = zipfile.ZipFile(name)
    except zipfile.error, e:
        print "Bad zipfile (from %r): %s" % (repo_url, e)
        return
    for n in z.namelist():
        dest = os.path.join(build_path, n)
        destdir = os.path.dirname(dest)

        if "Dockerfile" in dest:
            dockerfile_path = destdir

        if not os.path.exists(destdir):
            os.makedirs(destdir)

        if not os.path.isdir(dest):
            data = z.read(n)
            f = open(dest, 'w')
            f.write(data)
            f.close()

    z.close()
    os.unlink(name)

    # replace any service parmeters variables in body of any files 
    apply_templates_in_dir(dockerfile_path, service_parmeters, build_path)

    return build_path

def get_image_name(image, version):

    return "%s:%s"%(image,version)

def parse_image_version(image_tag):

    image_parts = image_tag.split(':')

    return image_parts[1]




import argparse, json, os
from yac.lib.variables import get_variable, set_variable
from yac.lib.registry import set_private_registry, get_private_registry, clear_private_registry


def main():

    parser = argparse.ArgumentParser('Manage my yac registry')

    # optional args
    parser.add_argument('--set',      help='yac will prompt your for the connection params for a private registry',
                                      action='store_true')
    parser.add_argument('--show',     help='show the registry currently configured',
                                      action='store_true')
    parser.add_argument('--clear',    help='clear the private registry currently configured',
                                      action='store_true')    
    
    args = parser.parse_args()       

    if args.set:

        redis_host = raw_input("Please input the name of the Redis host you would like to use for your yac private registry >> ")

        redis_port = raw_input("Please input the port the Redis host listens on >> ")
        private_registry = {"host": redis_host, "port": redis_port}

        set_private_registry(private_registry)

    if args.show:

        private_registry = get_private_registry()

        if private_registry:
            print json.dumps(private_registry, indent=4)
        else:
            print "No private registry is currently configured"

    if args.clear:

        print "Clearing private registry preferences currently in place"
        raw_input("Hit Enter to continue >> ")
        clear_private_registry()            





import shelve  # for local cache
import os, sys, time, datetime

from yac.lib.paths import get_config_path

class CacheError():
    def __init__(self, msg):
        self.msg = msg
            
# set expiration in ms
def set_cache_value_ms(key_name, value, expiration_ms=""):

    # save value to shelve db
    yac_db = _get_cache()

    if not expiration_ms:
        # set expiration to next year
        expiration_dt = datetime.date.today() + datetime.timedelta(days=365)
        # convert to ms
        expiration_ms = int(expiration_dt.strftime("%s")) * 1000

    yac_db[key_name] = {"value": value, "expiration_ms": expiration_ms}

# set expiration using a datetime object
def set_cache_value_dt(key_name, value, expiration_dt=""):

    # save value to shelve db
    yac_db = _get_cache()

    if not expiration_dt:
        # set expiration to next year
        expiration_dt = datetime.date.today() + datetime.timedelta(days=365)

    # convert to ms
    expiration_ms = int(expiration_dt.strftime("%s")) * 1000

    yac_db[key_name] = {"value": value, "expiration_ms": expiration_ms}

def get_cache_value(key_name, default_value={}):  

    cache_db = _get_cache()

    # current time in ms
    time_ms = int(time.time() * 1000)

    if (key_name in cache_db and time_ms < cache_db[key_name]["expiration_ms"]):

        # convert the expiration to millis        
        return cache_db[key_name]['value']
    else:
        return default_value

def delete_cache_value(key_name):   

    cache_db = _get_cache()

    if key_name in cache_db:
        cache_db.pop(key_name)

def get_cache_keys():

    cache_db = _get_cache()

    return cache_db.keys()        

def _get_cache():

    # cache should be saved under the user's home directory
    home = os.path.expanduser("~")
    db_home = os.path.join(home,'.yac')

    if not os.path.exists(db_home):
        os.makedirs(db_home)

    yac_db_path = os.path.join( db_home,'yac_cache')

    cache_db = shelve.open(yac_db_path)

    return cache_db



import re, boto3, os, json, jmespath, sys, argparse
from yac.lib.variables import get_variable, set_variable

# state is persisted in a json file in s3
def load_state_s3(s3_bucket, s3_path, service_alias):

    state = {}

    s3 = boto3.resource('s3')

    state_filename = get_state_filename(service_alias)

    state_file_local_path_full = get_state_local_path(s3_path,service_alias)

    s3_file_path = "%s/%s"%(s3_path,state_filename)

    if state_exists(s3_bucket, s3_path):

        # pull file down to the local dir
        s3.meta.client.download_file(s3_bucket, 
                                     s3_file_path, 
                                     state_file_local_path_full)

        # pull contents into a dictionary
        if os.path.exists(state_file_local_path_full):
            with open(state_file_local_path_full) as file_arg_fp:
                file_contents = file_arg_fp.read()
                state = json.loads(file_contents)

    return state

def state_exists(s3_bucket,s3_path):

    object_exists = False

    s3 = boto3.resource('s3')
    bucket = s3.Bucket(s3_bucket)
    objs = list(bucket.objects.filter(Prefix=s3_path))
    if len(objs) > 0 and objs[0].key == s3_path:
        object_exists = True

    return object_exists

def save_state_s3(s3_bucket, s3_path, service_alias, state):

    s3 = boto3.resource('s3')

    state_file_local_path_full = get_state_local_path(s3_path, service_alias)

    # write state dictionary to the file
    state_str = json.dumps(state, indent=2)

    # make sure dir exists
    state_file_local_path = os.path.basename(state_file_local_path_full)
    
    if not os.path.exists(state_file_local_path):
        os.makedirs(state_file_local_path)

    with open(state_file_local_path_full, 'w') as file_arg_fp:
        file_arg_fp.write(state_str)
        
    state_filename = get_state_filename(service_alias)            
    s3_file_path = "%s/%s"%(s3_path,state_filename)

    # write file to s3
    s3.meta.client.upload_file(state_file_local_path_full, 
                               s3_bucket,
                               s3_file_path)

def get_state_filename(service_alias):

    state_filename = "%s-state.json"%(service_alias)

    return state_filename

def get_state_local_path(s3_path,service_alias):

    home = os.path.expanduser("~")
    state_file_local_path = os.path.join(home,'.yac',s3_path)
    state_filename = get_state_filename(service_alias)
    state_file_local_path_full = os.path.join(state_file_local_path,state_filename)

    # make sure local path exists
    if not os.path.exists(state_file_local_path):
        os.makedirs(state_file_local_path)

    return state_file_local_path_full


import requests, jmespath

# for the registry find, find the latest version of the app input, in the container org input
def get_latest_version(app, container_org, registry_url='https://registry.hub.docker.com'):

    # endpoint use to determine the latest version in the registry input for this app
    endpoint_uri = "/v2/repositories/%s/%s/tags"%(container_org,app)

    # hit the endpoint
    endpoint_response = requests.get(registry_url + endpoint_uri) 

    # use jmespath to extract just the version value for each
    versions = jmespath.search("results[*].name",endpoint_response.json())

    # jmespath leaves a weird prefix on each result. remove by str conversion
    latest_version = "not found!"

    if versions:

        versions = [str(i) for i in versions]

        latest_version = ""

        if (versions and len(versions)>=1):

            if 'latest' in versions:

                latest_version = "latest"
            
            else:
                
                # sort and grab the last value in the resulting list (the latest version)
                versions.sort()
                latest_version = versions[-1]

    return str(latest_version)

def get_app_version(app_alias, my_service_descriptor):

    app_version = ""

    # the version would be in the image for this app
    search_str = "task-definition.containerDefinitions[?name=='%s'].image"%app_alias
    images_found = jmespath.search(search_str, my_service_descriptor)

    if images_found and len(images_found)==1:

        # grab the version from the tail end of the image label
        app_version = str(images_found[0]).split(':')[-1]

    return app_version    

# @param version1, a string, e.g. 1.2.1
# @param version2, a string
# @returns version 1 or version2, whichever is the latest
def return_latest(version1, version2):

    v1, v2 = version1.split("."), version2.split(".")
    
    if len(v1) > len(v2):
        v2 += ['0' for _ in xrange(len(v1) - len(v2))]
    elif len(v1) < len(v2):
        v1 += ['0' for _ in xrange(len(v2) - len(v1))]
    
    latest = ""
    i = 0
    while i < len(v1):
        if (v1[i].isdigit() and v2[i].isdigit() and
            int(v1[i]) > int(v2[i]) ):
            latest = version1
            break
        elif (v1[i].isdigit() and v2[i].isdigit() and
            int(v1[i]) < int(v2[i]) ):
            latest = version2
            break
        else:
            i += 1
    
    return latest

# @param version1, a string, e.g. 1.2.1
# @param version2, a string
# @returns True if both versions are the same, False o.w.
def is_same(version1, version2):

    v1, v2 = version1.split("."), version2.split(".")
    
    if len(v1) > len(v2):
        v2 += ['0' for _ in xrange(len(v1) - len(v2))]
    elif len(v1) < len(v2):
        v1 += ['0' for _ in xrange(len(v2) - len(v1))]
    
    same = True
    i = 0
    while i < len(v1):
        if (v1[i].isdigit() and v2[i].isdigit() and
            int(v1[i]) == int(v2[i]) ):
            same = same and True
        else:
            same = same and False
        i += 1
    
    return same    

# @param version1, a string, e.g. 1.2.1
# @param version2, a string
# @returns True if version 1 is later than version2, False o.w.
def is_first_arg_latest(version1, version2):

    version = return_latest(version1, version2)    
    
    if version == version1:
        return True
    elif version == version2:
        return False
    else:

        return False        

def is_a_version(version):

    v1 = version.split(".")

    is_version = True
    i=0
    while i < len(v1):

        if not v1[i].isdigit():
            is_version = False
            break
        else:
            i += 1 

    return is_version    


#!/usr/bin/env python

import os, boto3, botocore, jmespath
from sqlalchemy import *
from sqlalchemy.orm import sessionmaker
from sqlalchemy_utils import database_exists
import copy

# describes how to connect to a DB
# used for inputs to most functions below
class DB():

    def __init__(self, 
                 db_name,
                 db_user,
                 db_pwd,
                 db_host,
                 db_port,
                 db_engine):

        self.configs = {}

        self.configs['db_name'] = db_name
        self.configs['db_user'] = db_user
        self.configs['db_pwd']  = db_pwd
        self.configs['db_host'] = db_host
        self.configs['db_port'] = db_port        
        self.configs['engine'] = db_engine

# get the driver corresponding to the os we are running on
def get_db_driver(this_os = os.name):

    # drivers that seem to work best include:
    # windows:  pyodbc
    # macos:    pymssql

    if this_os == 'posix':
        # for posix-compliant os like macos
        driver = 'pymssql'
    else:
        # for windows this one seems to work best
        driver = 'pyodbc'

    return driver

# get connection string for connecting to an ms-sql RDBMS 
# db is an instance of the DB class
def get_connection_string(db):
    connection = ""

    if db.configs['engine']=='mssql':
        driver = get_db_driver()
        connection = "mssql+%s"%driver
    elif db.configs['engine']=='postgres':
        # pg8000 is the only "pure-python" driver available for python 2.7
        connection = 'postgresql+pg8000'
        #connection = 'postgresql'

    connection_string = "%s://%s:%s@%s/%s" % (connection, 
                                        db.configs['db_user'], 
                                        db.configs['db_pwd'], 
                                        db.configs['db_host'], 
                                        db.configs['db_name'])

    return connection_string

# connect to database
# db is an instance of the DB class
def connect_to_db(db):
    return create_engine(get_connection_string(db))
    
# get an executable table (a table that alchemy commands can be run against)
# db is an instance of the DB class
def get_table(table_name, db):

    engine = connect_to_db(db)
    connection = engine.connect()
    metadata = MetaData(engine)

    # build table
    return Table(table_name, metadata, autoload=True)

def get_session(db):

    engine = connect_to_db(db)

    # create a configured "Session" class
    Session = sessionmaker(bind=engine)

    # create a Session
    session = Session()

def rename_rds_instance(old_name, new_name):

  # rename
  print 'renaming %s to %s'%(old_name, new_name)

  client = boto3.client('rds')
  response = client.modify_db_instance(
        DBInstanceIdentifier=old_name,
        ApplyImmediately=True,
        NewDBInstanceIdentifier=new_name)

# take all of the security groups assigned to instance_to_mimic_id and assign them to
# instance_id
def restore_security_group(instance_id, instance_to_mimic_id):
            
    try:
        client = boto3.client('rds')

        response = client.describe_db_instances(DBInstanceIdentifier=instance_to_mimic_id)

        security_group_groups = jmespath.search("DBInstances[*].VpcSecurityGroups[*].VpcSecurityGroupId", response)

        # the outer group should correspond to the unique instance specified, so should have 
        # length==1
        if len(security_group_groups)==1:

            security_group_ids = security_group_groups[0]
            client.modify_db_instance(DBInstanceIdentifier=instance_id,
                                      VpcSecurityGroupIds=security_group_ids)

    except botocore.exceptions.ClientError as e:

        print "Could not add security group(s) to %s. Error %s"%(instance_id,str(e))  

def delete_instance(instance_id):

    client = boto3.client('rds')

    response = client.delete_db_instance(
        DBInstanceIdentifier=instance_id,
        SkipFinalSnapshot=True)

def snapshot_exists(snapshot_name):

    client = boto3.client('rds')

    try:
        response = client.describe_db_snapshots(
            DBSnapshotIdentifier=snapshot_name)

        # should return exactly one snapshot
        return (len(response['DBSnapshots'])==1)

    except botocore.exceptions.ClientError as e:
        # print "Could not find rds snapshot. Error %s"%str(e)
        return False  

def instance_exists(instance_id):

    client = boto3.client('rds')

    try:
        response = client.describe_db_instances(
            DBInstanceIdentifier=instance_id)

        # should return exactly one instance
        return (len(response['DBInstances'])==1)

    except botocore.exceptions.ClientError as e:
        # print "Could not find rds instance. Error %s"%str(e)
        return False         

# instance_id: id of instance to create from snapshot
# instance_to_mimic_id: id of instance with parameters to mimic (usually the original instance)
# snapshot_name: name of the snapshot to restore into instance_id
def restore_intance_from_snapshot(instance_id, instance_to_mimic_id, snapshot_name):

    try:

        # start a new client session
        client = boto3.client('rds')

        response = client.describe_db_instances(DBInstanceIdentifier=instance_to_mimic_id)

        if len(response['DBInstances'])==1:

            instance_params = response['DBInstances'][0]

            # restore the snapshot into a new instance
            client.restore_db_instance_from_db_snapshot(DBInstanceIdentifier=instance_id,
                DBSnapshotIdentifier=snapshot_name,
                DBInstanceClass=instance_params['DBInstanceClass'],
                DBSubnetGroupName=instance_params['DBSubnetGroup']['DBSubnetGroupName'],
                PubliclyAccessible=instance_params['PubliclyAccessible'],
                MultiAZ=instance_params['MultiAZ']) 

    except botocore.exceptions.ClientError as e:

        print "Could not find instance to mimic. Error %s"%str(e)

# db exists
def db_exists(db, db_name):

    db2 = copy.deepcopy(db)
    db2.configs['db_name'] = db_name
    connection_str = get_connection_string(db2)

    return database_exists(connection_str)

# create a DB
def create_db(db, create_db_sql):

    connection_str = get_connection_string(db)

    engine = create_engine(connection_str,
                           isolation_level='AUTOCOMMIT')
    
    conn = engine.connect()
    
    conn.execute(create_db_sql)
    conn.close()  
    
# drop a DB
def drop_db(db, drop_db_sql):

    connection_str = get_connection_string(db)

    engine = create_engine(connection_str,
                           isolation_level='AUTOCOMMIT')
    
    conn = engine.connect()
    
    conn.execute(drop_db_sql)
    conn.close()


import argparse, os, inspect, jmespath, json
from colorama import Fore, Style
from yac.lib.service import get_service, get_service_parmeters, get_service_alias
from yac.lib.stack import get_stack_templates, get_stack_name, get_ec2_ips, deploy_stack_files
from yac.lib.naming import set_namer
from yac.lib.params import get_service_params, NULL_PARAMS, INVALID_PARAMS
from yac.lib.vpc import get_vpc_prefs 
from yac.lib.variables import get_variable,set_variable
from yac.lib.cache import get_cache_value, set_cache_value_dt

def is_valid_file(parser, arg):
    if not os.path.exists(arg):
        parser.error("The key file %s does not exist!" % arg)
    else:
        return arg

def main():

    path_help =   'path to the private key file for this service'
    user_help =   'the user to login as'
    public_help = 'connect using the public IP address?'

    parser = argparse.ArgumentParser('ssh into a stack ec2 instance')

    parser.add_argument('servicefile', help='location of the Servicefile (registry key or abspath)')
    parser.add_argument('--path', help=path_help + ' (defaults to previous)',
                                       type=lambda x: is_valid_file(parser, x))
    parser.add_argument('-u','--user', help=user_help + ' (defaults to previous')
    parser.add_argument('-s','--search',help='search string, to support multiple EC2 instances in the same auto-scaling group')
    parser.add_argument('--public',     help=public_help + ' (defaults to previous)', 
                                        action='store_true')
    parser.add_argument('-a',
                        '--alias',  help='service alias for the stack currently supporting this service (deault alias is per Servicefile)')
    parser.add_argument('-p',
                        '--params',  help='path to a file containing additional, static, service parameters (e.g. vpc params, of service constants)') 

    # pull out args
    args = parser.parse_args()

     # determine service defintion, complete service name, and service alias based on the args
    service_descriptor, service_name, servicefile_path = get_service(args.servicefile) 

    # abort if service_descriptor was not loaded successfully
    if not service_descriptor:
        print("The Servicefile input does not exist locally or in registry. Please try again.")
        exit()

    # get vpc preferences in place
    vpc_prefs = get_vpc_prefs()

    # set the resource namer to use with this service
    set_namer(service_descriptor, servicefile_path, vpc_prefs)

    # get the alias to use with this service
    service_alias = get_service_alias(service_descriptor,args.alias)

    # determine service params based on the params arg
    service_params_input, service_params_name = get_service_params(args.params)

    # abort if service params were not loaded successfully
    if args.params and service_params_name == NULL_PARAMS:
        print("The service params specified do not exist locally or in registry. Please try again.")
        exit()
    elif service_params_name == INVALID_PARAMS:
        print("The service params file specified failed validation checks. Please try again.")
        exit()

     # get the service parameters for use with yac-ref's in service templates
    service_parmeters = get_service_parmeters(service_alias, service_params_input, 
                                              service_name, service_descriptor,
                                              servicefile_path, vpc_prefs)

    # get the name of the stack that corresponds to this app and env
    stack_name = get_stack_name(service_parmeters)

    # should we connect using a public ip addr?
    use_public = get_use_public(service_alias, args.public)

    # get the IP address of the ec2 instance for this stack
    ec2_ips = get_ec2_ips(stack_name, args.search, use_public)

    if (ec2_ips and len(ec2_ips)==1):

        ec2_ip = ec2_ips[0]

        # get the user to ssh in as
        ssh_user = get_ssh_user(service_alias, user_help, args.user)

        # get the path to the ssh key for this service
        key_path = get_key_path(service_alias, path_help, args.path)

        # prepare ssh command
        ssh_cmd = "ssh -i %s %s@%s"%(key_path,ssh_user,ec2_ip)

        # add ingress on port 22 to the SG associated with the instance

        # ssh in
        os.system(ssh_cmd)

        print "future versions will remove 22 ingress here..."

    elif (ec2_ips and len(ec2_ips)>1):

        print 'Found multiple EC2 address for the %s app in the %s env. Consider adding a search string.'%(args.app, args,env)

    else:

        print 'Could not find address for the %s app in the %s env'%(args.app, args,env)

def get_use_public(service_alias, arg_input):

    cache_key = "%s:arg:use-public"

    use_public = get_cached_arg_bool('public', arg_input, cache_key)

    return use_public

def get_key_path(service_alias, arg_help, arg_input):

    cache_key = "%s:arg:ssh-key"

    validation_fxn = "os.path.exists"

    key_path = get_cached_arg_val('path', arg_input, cache_key, arg_help, validation_fxn)

    return key_path

def get_ssh_user(service_alias, arg_help, arg_input):

    cache_key = "%s:arg:ssh-user"

    ssh_user = get_cached_arg_val('path', arg_input, cache_key, arg_help)

    return ssh_user

def get_cached_arg_val(arg_name, arg_input, cache_key, input_help, validation_fxn=""):

    cached_arg_val = ""

    if arg_input:
        cached_arg_val = arg_input
    else:
        # see if the arg is in the user's local cache
        cached_arg_val = get_cache_value(cache_key, '')

        if not cached_arg_val:

            print "The %s input is not yet cached locally."%arg_name
            print "Arg description: %s."%arg_help

            input_prompt_msg = "Please enter a value for %s >> "%arg_name

            if validation_fxn:
                while not validation_fxn(cached_arg_val):
                    cached_arg_val = raw_input(input_prompt_msg)                
            else:
                cached_arg_val = raw_input(input_prompt_msg)

    # save pwd back to the local cache for future reference
    # with max timeout
    set_cache_value_dt(cache_key,
                       cached_arg_val)

    return cached_arg_val

def get_cached_arg_bool(arg_name, arg_input, cache_key):

    cached_arg_val = ""

    if arg_input:
        cached_arg_val = arg_input
    else:
        # see if the arg is in the user's local cache
        cached_arg_val = get_cache_value(cache_key, '')

    # save pwd back to the local cache for future reference
    # with max timeout
    set_cache_value_dt(cache_key,
                       cached_arg_val)

    return cached_arg_val



import os, imp
from yac.lib.intrinsic import apply_custom_fxn
from yac.lib.variables import get_variable,set_variable
from yac.lib.file import get_localized_script_path
from yac.lib.inputs import get_inputs

def get_task(task_name, service_parameters):

    task = {}
    tasks = get_variable(service_parameters,"tasks",{})

    if tasks:

        if task_name in tasks.keys():
            task = tasks[task_name]

    return task

def get_task_names(service_parameters):

    task_names = []

    tasks = get_variable(service_parameters,"tasks",{})

    if tasks:

        task_names = tasks.keys()

    return task_names

def get_task_script(task_name, service_parameters):

    task_script = ""
    tasks = get_variable(service_parameters,"tasks",{})

    if tasks and get_variable(tasks,task_name):

        task_script = get_variable(tasks,task_name)['script']

    return task_script

def get_task_inputs(task_name, service_parameters):

    inputs = {}
    conditional_inputs = {}

    tasks = get_variable(service_parameters,"tasks",{})

    if (tasks and get_variable(tasks,task_name) and 'inputs' in get_variable(tasks,task_name)):

        inputs = get_variable(tasks,task_name)['inputs']

    if (tasks and get_variable(tasks,task_name) and 'conditional-inputs' in get_variable(tasks,task_name)):

        conditional_inputs = get_variable(tasks,task_name)['conditional-inputs']

    return inputs, conditional_inputs

def run_task(task_name, service_parameters, stack_template):

    return_val = ""

    script_rel_path = get_task_script(task_name, service_parameters)

    inputs, conditional_inputs = get_task_inputs(task_name, service_parameters)

    script_path = get_localized_script_path(script_rel_path,service_parameters)

    if (not script_path or not os.path.exists(script_path)):

        print 'task %s executable does not exist'%script_path

    else:

        module_name = 'yac.lib.customizations'
        script_module = imp.load_source(module_name,script_path)

        # call the task_handler fxn in the task script
        return_val = script_module.task_handler(service_parameters,stack_template,inputs, conditional_inputs)
        
    return return_val

def handle_task_inputs(inputs,conditional_inputs):

    task_params = {}

    task_params = {"task-inputs": 
                    { "inputs": inputs,
                      "conditional-inputs": conditional_inputs
                    }}

    # prompt for inputs associated with this task                    
    get_inputs(task_params, "task-inputs")

    return task_params      